import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;

/** Panel holding the target image for warping. */
class WarpTargetPanel extends ImagePanel{
	private Laf icApp;
	private ICInternalFrame iFrame;
	private InternalFrameAdapter closeListener;
	private BufferedImage source, orig;
	private int origX, origY;	// coordinates of the source image in orignal image.
	private ImageShape warp;	// image shape for warped image.
	private Shape [] path;		// 4 target selectors.
	private XYPair [] sPath={new XYPair(0, 0), new XYPair(), new XYPair(), new XYPair()};	// 4 source selectors.
	private Circle drag;		// selector being dragged
	private boolean [] features={true, true, true};

	/** Constructor:
	 * @param bi Image to warp.
	 * @param orig Original image with the source image cut out.
	 * @param x X coordinate of the source in original image.
	 * @param y Y coordinate of the source in original image.
	 * @param icApp The main ImageCompositing application.
	 * @param iFrame Internal frame this panel is in.
	 */
	public WarpTargetPanel(BufferedImage bi, BufferedImage orig, int x, int y,
		Laf icApp, ICInternalFrame iFrame){
		this.icApp=icApp;
		this.iFrame=iFrame;
		this.orig=orig;

		setImage(orig);
		setState(MISC);

		// set up keyboard event to listen to Enter key.
		InputMap im = iFrame.getInputMap();
		ActionMap am = iFrame.getActionMap();
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), "down");

		// event handlers for keys, perform nudges to the selectors.
		am.put("down", new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				finilizeOperation();
			}
		});
		
		iFrame.addInternalFrameListener((closeListener=new InternalFrameAdapter(){
			public void internalFrameClosing(InternalFrameEvent e){
				if(finilizeOperation())
					WarpTargetPanel.this.iFrame.dispose();
			}
		}));
		
		source=bi;
		// source warp points.
		sPath[1].y=sPath[3].x=0;
		sPath[1].x=sPath[2].x=bi.getWidth()-1;
		sPath[2].y=sPath[3].y=bi.getHeight()-1;
		
		// warp-box.
		Color c=new Color(128, 128, 128);
		Shape [] tmp={
			new Circle(3, x, y, c),
			new Circle(3, x+bi.getWidth()-1, y, c),
			new Circle(3, x+bi.getWidth()-1, y+bi.getHeight()-1, c),
			new Circle(3, x, y+bi.getHeight()-1, c),
			null
		};
		tmp[4]=new WarpRect(((Circle)tmp[0]).center,
			((Circle)tmp[1]).center,
			((Circle)tmp[2]).center,
			((Circle)tmp[3]).center, c, 3);
		path=tmp;

		warp=new ImageShape(bi, x, y);
		origX=x;
		origY=y;
		iFrame.grabFocus();
	}
	
	/** Handles the button press.
	 * @param e Event object describing the event.
	 */
	public void actionPerformed(){
		XYPair rc1, rc2, rc3, rc4;
		// store the points again so they don't get changed.
		rc1=new XYPair(((Circle)path[0]).center.x, ((Circle)path[0]).center.y);
		rc2=new XYPair(((Circle)path[1]).center.x, ((Circle)path[1]).center.y);
		rc3=new XYPair(((Circle)path[2]).center.x, ((Circle)path[2]).center.y);
		rc4=new XYPair(((Circle)path[3]).center.x, ((Circle)path[3]).center.y);
		
		warp=new ImageShape(Paint.getWarp(
			source,
			sPath[0], sPath[1], sPath[2], sPath[3],
			rc1, rc2, rc3, rc4, features), ((Circle)path[0]).center.x-rc1.x,
			((Circle)path[0]).center.y-rc1.y);
	}

	/**
	 * Finalizes some operation this image panel was doing.
	 * @return true if finilization was completed, false otherwise.
	 */
	public boolean finilizeOperation(){
		int ret=JOptionPane.showConfirmDialog(icApp, "Apply the warp?");
		ImagePanel ip=new ImagePanel(getImage());
		if(ret==JOptionPane.YES_OPTION){
			// store the warp.
			ip.paintImage(warp.getImage(), warp.getTopLeft().x, warp.getTopLeft().y);
		}else if(ret==JOptionPane.NO_OPTION){
			// restore original image.
			ip.paintImage(source, origX, origY);
		}else{
			iFrame.grabFocus();
			return false;
		}
		iFrame.setDefaultCloseOperation(JInternalFrame.DISPOSE_ON_CLOSE);
		iFrame.setContentPane(ip);
		((ImagePanel)iFrame.getContentPane()).setState(icApp.getImageState());
		iFrame.removeInternalFrameListener(closeListener);
		icApp.setActive(true);
		return true;
	}

	/** Event handler for mouse presses.
	 * @param e Event object describing the mouse event.
	 */
	public void mousePressed(MouseEvent e) {
		super.mousePressed(e);
		int x=e.getX(), y=e.getY();
		drag=null;
		for(int i=0; i<path.length-1; i++){
			if(((Circle)path[i]).hitTest(x, y)){
				drag=(Circle)path[i];
				break;
			}
		}
	}

	/** Event handler for mouse releases.
	 * @param e Event object describing the mouse event.
	 */
	public void mouseReleased(MouseEvent e){
		super.mouseReleased(e);
		if(drag!=null)
			actionPerformed();
	}

	/** Event handler for mouse clicks.
	 * @param e Event object describing the mouse event.
	 */
	public void mouseClicked(MouseEvent e){
		super.mouseClicked(e);
		if(e.getClickCount()>1)
			finilizeOperation();
	}
	
	/** Event handler for the mouse move event.
	 * @param e Object describing the event.
	 */
	public void mouseMoved(MouseEvent e){
		super.mouseMoved(e);
		int x=e.getX(), y=e.getY();
		for(int i=0; i<path.length-1; i++){
			if(((Circle)path[i]).hitTest(x, y)){
				setCursor(Laf.getCursor("hand"));
				return;
			}
		}
		setCursor(Laf.getCursor("default"));
	}
	
	/** Event handler for the mouse drag event.
	 * @param e Object describing the event.
	 */
	public void mouseDragged(MouseEvent e){
		super.mouseDragged(e);
		if(drag==null) return;
		drag.center.x=e.getX();
		drag.center.y=e.getY();
		repaint();
	}
		
	/** Paints this component.
	 * @param g Graphics object to paint on.
	 */
	public void paintComponent(Graphics g){
		super.paintComponent(g);
		warp.paint(g);
		for(int i=0; path!=null && i<path.length; i++)
			path[i].paint(g);
	}
}

